/*
  ParsAI Team
  Source code written and assembled Ehsan Foroughi
  version 0.8.0 - 22 Aug 2000

    Ehsan Foroughi,     Mehran Motamed,
    Pooya Karimian,     Hamed  Hatami, 
             M.Hossein Fardad.

    ParseAI team,
    Sharif University
    Tehran, Iran

  works to be done :
    side must be changed of half time 
    max ball speed, max player speed
*/

#define _player_extern_
#include "player.h"

#include "headers.h"

void refresh_monitor(void)
{
  int i;
  char s[MAX_MONITOR_SIZE], st[MAX_MONITOR_SIZE];
  double angle;

  if (!MONITOR_ENABLED) return;
  
  sprintf(s,"%d %d", player.unum, world.time.clock);
  sprintf(st," 0 %d %.2f %.2f %.2f", (int)world.ball.correctness, world.ball.body_dir, world.ball.x, world.ball.y);
  strcat(s,st);
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) {
    sprintf(st," %d %d %.2f %.2f %.2f", i, (int)world.tmate[i].correctness, world.tmate[i].body_dir, world.tmate[i].x, world.tmate[i].y);
    strcat(s,st);
  };
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) {
    sprintf(st," -%d %d %.2f %.2f %.2f", i, (int)world.opp[i].correctness, world.opp[i].body_dir, world.opp[i].x, world.opp[i].y);
    strcat(s,st);
  };
  i = get_2ball.needed_turns();
  sprintf(st," 13 0 0.0 %.2f %.2f", estimate_ball_x(i), estimate_ball_y(i));
  strcat(s,st);
  sprintf(st," 14 0 0.0 %.2f %.2f", world.offside.our.line, world.offside.opp.line);
  strcat(s,st);
  angle=shoot2goal.best_dir();
  sprintf(st," 15 0 %.2f %.2f %.2f", angle, world.ball.x, world.ball.y);
  strcat(s,st);
  double dir = dir_from_me(52,0);
  if (free_dist_cone(dir)>5) angle=dir;
  else angle = normalize_angle(dir+((free_dist_cone(normalize_angle(dir+45))>free_dist_cone(normalize_angle(dir-45)))?45:-45));
  sprintf(st," 16 0 %.2f %.2f %.2f", angle, world.ball.x, world.ball.y);
  strcat(s,st);

  if (LOG_SEND) logmsg(s);
  send_message(client.monitor_sock,s);
}

void debug(void)
{
  char s[1000];
  double vx, vy, px, py, ex, ey;
  double va, vd;

  if (!DEBUG_MODE) return;

  px = world.ball.x - world.body.loc.x;
  py = world.ball.y - world.body.loc.y;
  vx = world.ball.vx - world.body.loc.vx;
  vy = world.ball.vy - world.body.loc.vy;
  ex = px/dist_2me(world.ball);
  ey = py/dist_2me(world.ball);

  vd = vx*ex + vy*ey;
  va = (-vx*ey + vy*ex)/dist_2me(world.ball)*(180/M_PI);

  sprintf(s,"time %d %.2f = %.2f %.2f = %.2f", world.time.clock, 
          world.ball.vx, va, world.ball.vx, vd);  
  logmsg(s);
}

void refresh_time_dependents(void)
{
  update_world_bytime();
  update_skills_bytime();
  world.lasttime = world.time;
  logobjstates();
  refresh_monitor();
  debug();
}

bool timechanged(void)
{
  return (timeval(world.lasttime)<timeval(world.time));
}

void init_coach_world(void)
{
  int i;
  init_params();

  arrange.read_arrange();

  world.time.clock = 0;
  world.time.stopped = 0;

  if (player.playmode == PM_before_kick_off) world.time_running = false;
  else world.time_running = true;
  world.ball.x = 0;
  world.ball.y = 0;
  world.ball.vx = 0;
  world.ball.vy = 0;
  world.ball.correctness = 100;
  world.ball.update_time = world.time;
  world.body.loc.body_dir = (player.side==S_LEFT)?0:180;
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) {
    world.tmate[i].x = arrange.begin_position[i][(player.side==S_LEFT)?0:1][0];
    world.tmate[i].y = arrange.begin_position[i][(player.side==S_LEFT)?0:1][1];
    world.tmate[i].vx = 0;
    world.tmate[i].vy = 0;
    world.tmate[i].correctness = 100;
    world.tmate[i].update_time = world.time;
    world.opp[i].correctness = 0;
    world.tmate[i].stamina = STAMINA_MAX;
    world.opp[i].stamina = STAMINA_MAX;
  };
  sensor.body.time = 0;
  sensor.see.time = 0;
  sensor.hear.time = 0;
  logplayerstate();
}

void initialize_coach(bool blocking)
{
  InitInfo pl;
  client.sock = blocking?init_blocking_connection(client.server,client.port):
                         init_connection(client.server,client.port);
  if(client.sock.socketfd == -1) {
    fprintf(stderr, "Error :\n  cannot connect to server\n");
    exit(1);
  };

  pl = send_com_init_online_coach(&(client.sock), client.teamname, SERVER_VERSION);
  
  player.unum = pl.unum;
  player.side = pl.side;
  player.playmode = pl.playmode;
  player.playmodestr = pl.playmodestr;
  player.idtype = 0;
  
  if (player.side == S_UNKNOWN) {
       fprintf(stderr, "Can't init coach (side unknown)\n");
       exit(2);
  };

  if (LOGIT) {
    sprintf(client.log.filename, "%s_coach.log", client.teamname);
    client.log.file = fopen(client.log.filename,"w");
    fprintf(client.log.file, "\nOnline coach initiated successfully: \n side : %d\n playmode : %s\n\n",
            player.side,  player.playmodestr);
  }
  fprintf(client.log.file, "\nOnline coach initiated successfully: \n side : %d\n playmode : %s\n\n",
          player.side,  player.playmodestr);

  init_coach_world();
  
  if (MONITOR_ENABLED) {
    client.monitor_sock = init_connection(MONITOR_ADDR,MONITOR_PORT);
    if(client.monitor_sock.socketfd == -1) {
      fprintf(stderr, "Error :\n  cannot connect to monitor\n");
      exit(3);
    };
    printf("\n connected to monitor successfully");
    logmsg("\n connected to monitor successfully");
  };
}

void leavegame(void)
{
  send_com_bye(&(client.sock));
  close_connection(client.sock);
  if (LOGIT) fclose(client.log.file);
  exit(0);
}

// - *************************************************************************

void coach_new_turn()
{
  static int last_say_time=-1;
  static int last_time=-1;
  static int st = 0;
  if (kick_play_mode()) st++;
  else st=0;
/*  if (((last_say_time<world.time.clock)&&
       (last_time==world.time.clock))||
      (kick_play_mode())) {*/
  if (st==3) {
    coach_say_world_model();
    last_say_time = world.time.clock;
  }
  last_time = world.time.clock;
}

void process_received_msg(char *msg)
{
  ReceivedInfo received;

  logrcvdmsg(msg);
  scan_info(msg, &received, client.teamname);
  if (received.infotype == SensorT)
    update_world_bysensors(received.info.sensor);
  else if (received.infotype == ParamT)
    get_params(received.info.param);
  else if (received.infotype == FullStateT)
    get_fullstate_info(received.info.fullstate);
  else if (received.infotype == SeeGlobalT)
    get_seeglobal_info(received.info.see_global);
  else if (received.infotype == ErrorT)
    get_server_error(received.info.error_msg);
  if (client.role==COACH) {
    coach_new_turn();
  } else {
    if (timechanged())
      refresh_time_dependents();
  };
}

void coach_main_loop(void)
{
  char msg[MAIN_BUFFER_SIZE];

  while(true) {
    if (world.time.clock >= GAME_CYCLES) leavegame();
    alarm(AUTO_TERMINATE_SEC);  // Auto Terminate After (some) Seconds of not receiving data
    fflush(client.log.file);
    // Receive info from server
    if (!receive_message(&(client.sock), msg, MAIN_BUFFER_SIZE)) continue;

    process_received_msg(msg);
  };
}

int main(int argc, char **argv)
{
  client.server = DEFAULTSERVER;
  client.port = DEFAULTCOACHPORT;
  client.role = COACH;

  switch (argc) {
   case 4:
    client.port = atoi(argv[3]);
    client.server = argv[2];
   break;
   case 3:
    client.server = argv[2];
   break;
   case 2:
   break;
   default:
    printf("usage : %s TEAMNAME [HOST [PORT]]\n", argv[0]);
    exit(0);
   break;
  };

  strcpy(client.teamname, argv[1]);

  initialize_coach(true);

  eye(true);
  coach_main_loop();

  leavegame();
  return(0);
}
