/*
 *	Header:
 *	File: position.c
 *	Author: Noda Itsuki
 *      Modified by : Hamed Hatami ( Sharif University of Tech. )
 *	Date: 2000/08/1
 *	EndHeader:
 */

#include <stdio.h>
#include <math.h>

#include "libsclient.h"

#define Sin(x) sin(Deg2Rad(x))
#define Cos(x) cos(Deg2Rad(x))

Bool estimate_current_pos(SeeInfo *sinf, Side side, PosState *pstate)
{
  int i ;
  int out_of_field = FALSE ;
  double dist = 2.0*PITCH_LENGTH ;
  double line_dir, obj_dir ;
  double tmp ;
  SeenObj *line = (SeenObj *)NULL, *flag = (SeenObj *)NULL ;
  SeenObj *goal = (SeenObj *)NULL, *obj ;
  Pos obj_pos ;

  for(i = 0; i < sinf->n_obj; i++) {
    if (sinf->obj[i].type == OT_Flag) {
      if (!sinf->obj[i].inViewConeP)
	continue ;
      if (Abs(dist) > Abs(sinf->obj[i].dist)) {
	flag = &(sinf->obj[i]) ;
	dist = sinf->obj[i].dist ;
      }
    }
    else if (sinf->obj[i].type == OT_Line) {
      line = &(sinf->obj[i]) ;
      if (i != sinf->n_obj - 1 && sinf->obj[i+1].type == OT_Line)
	out_of_field = TRUE ;
      break ;
    }
    else if (sinf->obj[i].type == OT_Goal)
      goal = &(sinf->obj[i]) ;
  }

  if (line == (SeenObj *)NULL)
    return FALSE ;

  if (flag != (SeenObj *)NULL) {
    obj = flag ;
    for(i=0; ; i++) {
      if (Flags[i].name == FP_Unknown)
	return FALSE ;
      if (Flags[i].name == flag->id.fpos) {
	obj_pos.x = Flags[i].pos.x ;
	obj_pos.y = Flags[i].pos.y ;
	break ;
      }
    }
  }
  else if (goal != (SeenObj *)NULL) {
    obj = goal ;
    if (goal->id.side == S_LEFT)
      obj_pos.x = -PITCH_LENGTH/2.0 ;
    else if (goal->id.side == S_RIGHT)
      obj_pos.x = PITCH_LENGTH/2.0 ;
    else
      return FALSE ;
    obj_pos.y = 0.0 ;
  }
  else
    return FALSE ;

  line_dir = Deg2Rad(line->dir) ;

  switch (line->id.lpos) {
  case LP_T :
    if (out_of_field)  {
      pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
      pstate->pos.y = -PITCH_WIDTH/2.0 - Abs(line->dist * sin(line_dir)) ;
    }
    else {
      pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0) ;
      pstate->pos.y = -PITCH_WIDTH/2.0 + Abs(line->dist * sin(line_dir)) ;
    }

    if ((tmp = Pow(obj->dist) - Pow(obj_pos.y - pstate->pos.y)) < 0.0)
      tmp *= -1.0 ;

    obj_dir = obj->dir + pstate->dir ;
    if (obj_dir > 180.0)
      obj_dir -= 360.0 ;
    else if (obj_dir < -180.0)
      obj_dir += 360.0 ;

    if (Abs(obj_dir) < 90.0)
      pstate->pos.x = obj_pos.x - sqrt(tmp) ;
    else
      pstate->pos.x = obj_pos.x + sqrt(tmp) ;
    break ;

  case LP_B :
    if (out_of_field) {
      pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0) ;
      pstate->pos.y = PITCH_WIDTH/2.0 + Abs(line->dist * sin(line_dir)) ;
    }
    else {
      pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
      pstate->pos.y = PITCH_WIDTH/2.0 - Abs(line->dist * sin(line_dir)) ;
    }

    if ((tmp = Pow(obj->dist) - Pow(obj_pos.y - pstate->pos.y)) < 0.0)
      tmp *= -1.0 ;

    obj_dir = obj->dir + pstate->dir ;
    if (obj_dir > 180.0)
      obj_dir -= 360.0 ;
    else if (obj_dir < -180.0)
      obj_dir += 360.0 ;

    if (Abs(obj_dir) < 90.0)
      pstate->pos.x = obj_pos.x - sqrt(tmp) ;
    else
      pstate->pos.x = obj_pos.x + sqrt(tmp) ;
    break ;

  case LP_L :
    if (out_of_field) {
      pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0) ;
      pstate->pos.x = -PITCH_LENGTH/2.0 - Abs(line->dist * sin(line_dir)) ;
    }
    else {
      pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0) ;
      pstate->pos.x = -PITCH_LENGTH/2.0 + Abs(line->dist * sin(line_dir)) ;
    }

    if ((tmp = Pow(obj->dist) - Pow(obj_pos.x - pstate->pos.x)) < 0.0)
      tmp *= -1.0 ;

    obj_dir = obj->dir + pstate->dir ;
    if (obj_dir > 180.0)
      obj_dir -= 360.0 ;
    else if (obj_dir < -180.0)
      obj_dir += 360.0 ;

    if (obj_dir > 0.0)
      pstate->pos.y = obj_pos.y - sqrt(tmp) ;
    else
      pstate->pos.y = obj_pos.y + sqrt(tmp) ;
    break ;

  case LP_R :
    if (out_of_field) {
      pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0) ;
      pstate->pos.x = PITCH_LENGTH/2.0 + Abs(line->dist * sin(line_dir)) ;
    }
    else {
      pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0) ;
      pstate->pos.x = PITCH_LENGTH/2.0 - Abs(line->dist * sin(line_dir)) ;
    }

    if ((tmp = Pow(obj->dist) - Pow(obj_pos.x - pstate->pos.x)) < 0.0)
      tmp *= -1.0 ;

    obj_dir = obj->dir + pstate->dir ;
    if (obj_dir > 180.0)
      obj_dir -= 360.0 ;
    else if (obj_dir < -180.0)
      obj_dir += 360.0 ;

    if (obj_dir > 0.0)
      pstate->pos.y = obj_pos.y - sqrt(tmp) ;
    else
      pstate->pos.y = obj_pos.y + sqrt(tmp) ;
    break ;

  case LP_Unknown :
    return FALSE ;
  }

  if (side == S_RIGHT) {
    pstate->dir += ((pstate->dir > 0) ? -180.0 : 180.0) ;
    pstate->pos.x *= -1.0 ;
    pstate->pos.y *= -1.0 ;
  }
		
  return TRUE ;
}

Bool estimate_current_pos2(SeeInfo *sinf, Side side, PosState *pstate)
{
     int i ;
     int out_of_field = FALSE ;
     SeenObj *line = (SeenObj *)NULL ;
     double sum_x = 0.0, sum_y = 0.0 , n_x = 0.0, n_y = 0.0;

     for(i = 0; i < sinf->n_obj; i++) {
	  if (sinf->obj[i].type == OT_Line) {
	       if(line == NULL) {
		    line = &(sinf->obj[i]) ;
	       } else {
		    out_of_field = TRUE ;
	       }
	  }
     }
     if (line == (SeenObj *)NULL)
	  return FALSE ;

     if (line->dist < 0.000001)
         line->dist = 0.000001;
     switch (line->id.lpos) {
     case LP_T :
	  if (out_of_field)  {
	       double ey, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	       ey = -PITCH_WIDTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_y += ey * w ; n_y += w ;
	  } else {
	       double ey, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0) ;
	       ey = -PITCH_WIDTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_y += ey * w ; n_y += w ;
	  }
	  break ;
     case LP_B :
	  if (out_of_field) {
	       double ey, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0) ;
	       ey = PITCH_WIDTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_y += ey * w ; n_y += w ;
	  } else {
	       double ey, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	       ey = PITCH_WIDTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_y += ey * w ; n_y += w ;
	  }
	  break ;
     case LP_L :
	  if (out_of_field) {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0) ;
	       ex = -PITCH_LENGTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_x += ex * w ; n_x += w ;
	  } else {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0) ;
	       ex = -PITCH_LENGTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_x += ex * w ; n_x += w ;
	  } 
	  break ;
     case LP_R :
	  if (out_of_field) {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0) ;
	       ex = PITCH_LENGTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_x += ex * w ; n_x += w ;
	  } else {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0) ;
	       ex = PITCH_LENGTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
	       w = (1.0/Abs(line->dist)) ;
	       sum_x += ex * w ; n_x += w ;
	  }
	  break ;
     case LP_Unknown :
	  return FALSE ;
     }

     for(i = 0 ; i < sinf->n_obj ; i++) {
	  Pos obj_pos ;
	  double obj_dir, obj_dist ;
	  double ex, ey, w ;
	  SeenObj *obj ;
	  obj = &(sinf->obj[i]) ;

	  if(!(obj->inViewConeP)) continue ;
	  if(obj->type == OT_Flag) {
	       obj_pos.x = Flags[obj->id.fpos].pos.x ;
	       obj_pos.y = Flags[obj->id.fpos].pos.y ;
	  } else if(obj->type == OT_Goal) {
	       if(obj->id.side == S_LEFT) 
		    obj_pos.x = -PITCH_LENGTH/2.0 ;
	       else 
		    obj_pos.x =  PITCH_LENGTH/2.0 ;
	       obj_pos.y = 0.0 ;
	  } else {
	       continue ;
	  }

	  obj_dir = obj->dir + pstate->dir ;
	  obj_dist = obj->dist ;

          if (obj_dist < 0.000001)
              obj_dist = 0.000001;

	  ex = obj_pos.x - obj_dist * Cos(obj_dir) ;
	  ey = obj_pos.y - obj_dist * Sin(obj_dir) ;
	  w = 1.0 / Abs(obj_dist) ;
	  sum_x += w * ex ; n_x += w ;
	  sum_y += w * ey ; n_y += w ;
     }

     if(n_x + n_y <= 0.0) return FALSE ;

     pstate->pos.x = sum_x / n_x ;
     pstate->pos.y = sum_y / n_y ; 

     if (side == S_RIGHT) {
	  pstate->dir += ((pstate->dir > 0) ? -180.0 : 180.0) ;
	  pstate->pos.x *= -1.0 ;
	  pstate->pos.y *= -1.0 ;
     }
     return TRUE ;
}

#define square(x) ((x) * (x))

/* $B2~A1$NM>CO$"$j(B */
double error(double dist, double dir_err) {
  double dist_err;

  if (dist <= 14.0)
    dist_err = 0.1;
  else 
    dist_err = (dist - 14.0) / 200.0 + 0.1;

  dir_err = 1.0 - Cos(dir_err);

  return dist * dir_err + 1.0 * dist_err + dist_err * dir_err;
}

#define ERR_MAX 100.0

Bool estimate_current_pos3(SeeInfo *sinf, Side side, PosState *pstate)
{
     int i , j;
     int out_of_field = FALSE ;
     SeenObj *line = (SeenObj *)NULL ;
     Bool inViewConePp = FALSE;
    
     Pos err;
     
     err.x = ERR_MAX;
     err.y = ERR_MAX;     

     for(i = 0; i < sinf->n_obj; i++) {
	  if (sinf->obj[i].type == OT_Line) {
	       if(line == NULL) {
		    line = &(sinf->obj[i]) ;
	       } else {
		    out_of_field = TRUE ;
	       }
	  }
     }
     if (line == (SeenObj *)NULL)
	  return FALSE ;

     switch (line->id.lpos) {
     case LP_T :
	  if (out_of_field)  {
	       pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0);
               pstate->pos.y = -PITCH_WIDTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
               err.y = error(line->dist, 0.5);
	  } else {
	       pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0);
	       pstate->pos.y = -PITCH_WIDTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
               err.y = error(line->dist, 0.5);
	  }
	  break ;
     case LP_B :
	  if (out_of_field) {
	       pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0);
	       pstate->pos.y = PITCH_WIDTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
               err.y = error(line->dist, 0.5);
	  } else {
	       pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0);
	       pstate->pos.y = PITCH_WIDTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
               err.y = error(line->dist, 0.5);
	  }
	  break ;
     case LP_L :
	  if (out_of_field) {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0);
	       pstate->pos.x = -PITCH_LENGTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
               err.x = error(line->dist, 0.5);
	  } else {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0);
	       pstate->pos.x = -PITCH_LENGTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
               err.x = error(line->dist, 0.5);
	  } 
	  break ;
     case LP_R :
	  if (out_of_field) {
	       pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0);
	       pstate->pos.x = PITCH_LENGTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
               err.x = error(line->dist, 0.5);
	  } else {
	       pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0);
	       pstate->pos.x = PITCH_LENGTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
               err.x = error(line->dist, 0.5);
	  }
	  break ;
     case LP_Unknown :
	  return FALSE ;
     }

     for(i = 0 ; i < sinf->n_obj ; i++) {
          double e;
	  Pos obj_pos ;
	  double obj_dir, obj_dist ;
	  SeenObj *obj ;
	  obj = &(sinf->obj[i]) ;

	  if(!(obj->inViewConeP)) {
            inViewConePp = TRUE;
            continue ;
	  }
	  if(obj->type == OT_Flag) {
	       obj_pos.x = Flags[obj->id.fpos].pos.x ;
	       obj_pos.y = Flags[obj->id.fpos].pos.y ;
	  } else if(obj->type == OT_Goal) {
	       if(obj->id.side == S_LEFT) 
		    obj_pos.x = -PITCH_LENGTH/2.0 ;
	       else 
		    obj_pos.x =  PITCH_LENGTH/2.0 ;
	       obj_pos.y = 0.0 ;
	  } else {
	       continue ;
	  }

	  obj_dir = obj->dir + pstate->dir ;
	  obj_dist = obj->dist ;

          e = error(obj_dist, 1.0);
          if (e < err.x) {
	    pstate->pos.x = obj_pos.x - obj_dist * Cos(obj_dir) ;
            err.x = e;
	  }
          if (e < err.y) {
	    pstate->pos.y = obj_pos.y - obj_dist * Sin(obj_dir) ;
	    err.y = e;
	  }
     }

     if (err.x == ERR_MAX || err.y == ERR_MAX) return FALSE;


     if (inViewConePp) {
       Pos tmp;
       double e;
       for(i = 0 ; i < sinf->n_obj ; i++) {
	  if(!(sinf->obj[i].inViewConeP)) {
            if (sinf->obj[i].type == OT_Flag) {
              tmp.x = pstate->pos.x + sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
              tmp.y = pstate->pos.y + sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
              for (j = sinf->obj[i].begin; j <= sinf->obj[i].end; j++) {
                if (sqrt((tmp.x - Flags[j].pos.x) * (tmp.x - Flags[j].pos.x) + (tmp.y - Flags[j].pos.y) * (tmp.y - Flags[j].pos.y)) < 5.0) {
                  e = error(sinf->obj[i].dist, 1.0);
                  if (e < err.x) {
		    pstate->pos.x = Flags[j].pos.x - sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
                 err.x = e;
		  }
                  if (e < err.y) {
		    pstate->pos.y = Flags[j].pos.y - sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
                 err.y = e;
		  }
		}
	      }
	    } else if (sinf->obj[i].type == OT_Goal) {
              tmp.x = pstate->pos.x + sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
              tmp.y = pstate->pos.y + sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
              if (sqrt((tmp.x - 52.5) * (tmp.x - 52.5) + (tmp.y) * (tmp.y)) < 5.0) {
		e = error(sinf->obj[i].dist, 1.0);
		if (e < err.x) {
		  pstate->pos.x = 52.5 - sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
                  err.x = e;
		}
		if (e < err.y) {
		  pstate->pos.y = -sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
                  err.y = e;
		}
	      } else if (sqrt((tmp.x + 52.5) * (tmp.x + 52.5) + (tmp.y) * (tmp.y)) < 5.0) {
		e = error(sinf->obj[i].dist, 1.0);
		if (e < err.x) {
		  pstate->pos.x = -52.5 - sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
                 err.x = e;
		}
		if (e < err.y) {
		  pstate->pos.y = -sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
                 err.y = e;
		}		
	      }
	    }
	  }
       }
     }

     /* printf("%f %f (%f %f)\n", pstate->pos.x, pstate->pos.y, err.x, err.y); */
     /* printf("%d (%f %f)\n", sinf->time, pstate->pos.x, pstate->pos.y); */

     if (side == S_RIGHT) {
	  pstate->dir += ((pstate->dir > 0) ? -180.0 : 180.0) ;
	  pstate->pos.x *= -1.0 ;
	  pstate->pos.y *= -1.0 ;
     }
     return TRUE ;
}

 
Bool estimate_current_pos4(SeeInfo *sinf, Side side, PosState *pstate, PosState *errstate)
{
     int i , j;
     int out_of_field = FALSE ;
     SeenObj *line = (SeenObj *)NULL ;
     Bool inViewConePp = FALSE;
    
     Pos err;
     
     err.x = ERR_MAX;
     err.y = ERR_MAX;     

     for(i = 0; i < sinf->n_obj; i++) {
	  if (sinf->obj[i].type == OT_Line) {
	       if(line == NULL) {
		    line = &(sinf->obj[i]) ;
	       } else {
		    out_of_field = TRUE ;
	       }
	  }
     }
     if (line == (SeenObj *)NULL)
	  return FALSE ;

     switch (line->id.lpos) {
     case LP_T :
	  if (out_of_field)  {
	       pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0);
               pstate->pos.y = -PITCH_WIDTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
               err.y = error(line->dist, 0.5);
	  } else {
	       pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0);
	       pstate->pos.y = -PITCH_WIDTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
               err.y = error(line->dist, 0.5);
	  }
	  break ;
     case LP_B :
	  if (out_of_field) {
	       pstate->dir = -line->dir - ((line->dir < 0) ? 180.0 : 0.0);
	       pstate->pos.y = PITCH_WIDTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
               err.y = error(line->dist, 0.5);
	  } else {
	       pstate->dir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0);
	       pstate->pos.y = PITCH_WIDTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
               err.y = error(line->dist, 0.5);
	  }
	  break ;
     case LP_L :
	  if (out_of_field) {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0);
	       pstate->pos.x = -PITCH_LENGTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
               err.x = error(line->dist, 0.5);
	  } else {
	       double ex, w ;
	       pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0);
	       pstate->pos.x = -PITCH_LENGTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
               err.x = error(line->dist, 0.5);
	  } 
	  break ;
     case LP_R :
	  if (out_of_field) {
	       pstate->dir = -line->dir - ((line->dir < 0) ? -90.0 : 90.0);
	       pstate->pos.x = PITCH_LENGTH/2.0 + Abs(line->dist * Sin(line->dir)) ;
               err.x = error(line->dist, 0.5);
	  } else {
	       pstate->dir = -line->dir - ((line->dir > 0) ? -90.0 : 90.0);
	       pstate->pos.x = PITCH_LENGTH/2.0 - Abs(line->dist * Sin(line->dir)) ;
               err.x = error(line->dist, 0.5);
	  }
	  break ;
     case LP_Unknown :
	  return FALSE ;
     }

     for(i = 0 ; i < sinf->n_obj ; i++) {
          double e;
	  Pos obj_pos ;
	  double obj_dir, obj_dist ;
	  SeenObj *obj ;
	  obj = &(sinf->obj[i]) ;

	  if(!(obj->inViewConeP)) {
            inViewConePp = TRUE;
            continue ;
	  }
	  if(obj->type == OT_Flag) {
	       obj_pos.x = Flags[obj->id.fpos].pos.x ;
	       obj_pos.y = Flags[obj->id.fpos].pos.y ;
	  } else if(obj->type == OT_Goal) {
	       if(obj->id.side == S_LEFT) 
		    obj_pos.x = -PITCH_LENGTH/2.0 ;
	       else 
		    obj_pos.x =  PITCH_LENGTH/2.0 ;
	       obj_pos.y = 0.0 ;
	  } else {
	       continue ;
	  }

	  obj_dir = obj->dir + pstate->dir ;
	  obj_dist = obj->dist ;

          e = error(obj_dist, 1.0);
          if (e < err.x) {
	    pstate->pos.x = obj_pos.x - obj_dist * Cos(obj_dir) ;
            err.x = e;
	  }
          if (e < err.y) {
	    pstate->pos.y = obj_pos.y - obj_dist * Sin(obj_dir) ;
	    err.y = e;
	  }
     }

     if (err.x == ERR_MAX || err.y == ERR_MAX) return FALSE;


     if (inViewConePp) {
       Pos tmp;
       double e;
       for(i = 0 ; i < sinf->n_obj ; i++) {
	  if(!(sinf->obj[i].inViewConeP)) {
            if (sinf->obj[i].type == OT_Flag) {
              tmp.x = pstate->pos.x + sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
              tmp.y = pstate->pos.y + sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
              for (j = sinf->obj[i].begin; j <= sinf->obj[i].end; j++) {
                if (sqrt((tmp.x - Flags[j].pos.x) * (tmp.x - Flags[j].pos.x) + (tmp.y - Flags[j].pos.y) * (tmp.y - Flags[j].pos.y)) < 5.0) {
                  e = error(sinf->obj[i].dist, 1.0);
                  if (e < err.x) {
		    pstate->pos.x = Flags[j].pos.x - sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
                 err.x = e;
		  }
                  if (e < err.y) {
		    pstate->pos.y = Flags[j].pos.y - sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
                 err.y = e;
		  }
		}
	      }
	    } else if (sinf->obj[i].type == OT_Goal) {
              tmp.x = pstate->pos.x + sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
              tmp.y = pstate->pos.y + sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
              if (sqrt((tmp.x - 52.5) * (tmp.x - 52.5) + (tmp.y) * (tmp.y)) < 5.0) {
		e = error(sinf->obj[i].dist, 1.0);
		if (e < err.x) {
		  pstate->pos.x = 52.5 - sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
                  err.x = e;
		}
		if (e < err.y) {
		  pstate->pos.y = -sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
                  err.y = e;
		}
	      } else if (sqrt((tmp.x + 52.5) * (tmp.x + 52.5) + (tmp.y) * (tmp.y)) < 5.0) {
		e = error(sinf->obj[i].dist, 1.0);
		if (e < err.x) {
		  pstate->pos.x = -52.5 - sinf->obj[i].dist * Cos(sinf->obj[i].dir + pstate->dir);
                 err.x = e;
		}
		if (e < err.y) {
		  pstate->pos.y = -sinf->obj[i].dist * Sin(sinf->obj[i].dir + pstate->dir);
                 err.y = e;
		}		
	      }
	    }
	  }
       }
     }

     /* printf("%f %f (%f %f)\n", pstate->pos.x, pstate->pos.y, err.x, err.y); */
     /* printf("%d (%f %f)\n", sinf->time, pstate->pos.x, pstate->pos.y); */

     if (err.x >= err.y) {
       errstate->pos.x = err.x;
       errstate->pos.y = err.x;
     } else {
       errstate->pos.x = err.y;
       errstate->pos.y = err.y;
     }
     errstate->dir = 0.5;

     if (side == S_RIGHT) {
	  pstate->dir += ((pstate->dir > 0) ? -180.0 : 180.0) ;
	  pstate->pos.x *= -1.0 ;
	  pstate->pos.y *= -1.0 ;
     }
     return TRUE ;
}

double quant(double x)
{
  return (0.1 * rint( exp(0.01 * rint( log(x) / 0.01 )) / 0.1));
}

void find_real_dist(double fake_dist,double * low,double * high)
{
  double i;
  *low=-1;
  *high=-1;
  for(i=fake_dist-0.5;i<fake_dist+2;i+=0.01){
    if ((*low==-1) && (fabs(quant(i)-fake_dist)<.001)){
      *low=i;
    }
    if ((*low!=-1) && (fabs(quant(i)-fake_dist)>=.001)){
      *high=i;
      break;
    }
  }
}

double point_weight(double x,double y,double low,double high,PosState obj)
{
  double angle;
  double distance;
  double weight=0;
  angle=Rad2Deg(atan2(obj.pos.y-y,obj.pos.x-x));
  distance=sqrt(Pow(obj.pos.y-y)+Pow(obj.pos.x-x));
  if (distance>high)
    weight+=distance-high;
  if (distance<low)
    weight+=low-distance;
  angle=fabs(angle-obj.dir);
  if (angle>180) 
    angle=360-angle;
  if (angle>1)
    weight+=(angle-1);  
  return weight;    
}

void swap(double *x,double *y)
{
  double temp;
  temp=*x;
  *x=*y;
  *y=temp;  
}

Bool estimate_current_pos5(SeeInfo *sinf, Side side, PosState *pstate)
{
     int i ,j;
     double dist_high[50],distance[50];
     PosState objs[50];
     double dist_low[50];
     int obj_num=0;
     double x,y,weight,k,t;
     double newx,newy,new_weight;
     double best_x,best_y,best_weight,temp;
     int turn_time;
     
     if (!estimate_current_pos3(sinf, side, pstate))
        if (!estimate_current_pos2(sinf, side, pstate))
	   if (!estimate_current_pos(sinf, side, pstate))
	     return false;     
	     
     if (side == S_RIGHT) {
	  pstate->dir += ((pstate->dir > 0) ? -180.0 : 180.0) ;
	  pstate->pos.x *= -1.0 ;
	  pstate->pos.y *= -1.0 ;
     }
     
    for(i = 0 ; i < sinf->n_obj ; i++) {
	  SeenObj *obj ;
	  obj = &(sinf->obj[i]) ;
	  if(!(obj->inViewConeP)) continue ;
	  if(obj->type == OT_Flag) {
	        objs[obj_num].pos.x= Flags[obj->id.fpos].pos.x ;
	        objs[obj_num].pos.y= Flags[obj->id.fpos].pos.y ;
	  } else if(obj->type == OT_Goal) {
	       if(obj->id.side == S_LEFT) 
		    objs[obj_num].pos.x = -PITCH_LENGTH/2.0 ;
	       else 
		    objs[obj_num].pos.x =  PITCH_LENGTH/2.0 ;
	       objs[obj_num].pos.y = 0.0 ;
	  } else {
	       continue ;
	  }

	  objs[obj_num].dir = obj->dir + pstate->dir;
	  if (objs[obj_num].dir>180) 
	    objs[obj_num].dir=objs[obj_num].dir-360;
	  if (objs[obj_num].dir<=-180) 
	    objs[obj_num].dir=objs[obj_num].dir+360;	  	    
	  distance[obj_num]=obj->dist;
	  for (j=obj_num-1;j>=0;j--)
	  {
	    if (distance[j]>distance[j+1]){
	      swap(&objs[j].pos.y,&objs[j+1].pos.y);
	      swap(&objs[j].pos.x,&objs[j+1].pos.x);
	      swap(&objs[j].dir,&objs[j+1].dir);
	      swap(&distance[j],&distance[j+1]);
	    }
	  }
          obj_num++;
     }     
     
     if (obj_num>5) 
        obj_num=5;
    for (i=0;i<=obj_num;i++)
      find_real_dist(distance[i],&dist_low[i],&dist_high[i]);	  
	
     x=best_x=pstate->pos.x;
     y=best_y=pstate->pos.y;
     weight=0;
     for (j=0;j<obj_num;j++){
       weight+=point_weight(x,y,dist_low[j],dist_high[j],objs[j]);
     }
     best_weight=weight;
     turn_time=0;
     for (i=0;i<100;i++)
       for (k=-1.0;k<2.0;k++)
         for (t=-1.0;t<2.0;t++)
	 {
           newx=x+k/25.0;
           newy=y+t/25.0;
           new_weight=0;
           for (j=0;j<obj_num;j++){
             new_weight+=point_weight(newx,newy,dist_low[j],dist_high[j],objs[j]);
           }
           if (new_weight<best_weight){
             best_x=newx;
	     best_y=newy;
	     best_weight=new_weight;
           }
           if (new_weight<weight){
             x=newx;
	     y=newy;	 
	     weight=new_weight;
	     turn_time=i;
           }
           if (turn_time+10<i){
             pstate->pos.x=best_x;
             pstate->pos.y=best_y;
             if (side == S_RIGHT) {
     	       pstate->dir += ((pstate->dir > 0) ? -180.0 : 180.0) ;
	       pstate->pos.x *= -1.0 ;
	       pstate->pos.y *= -1.0 ;
             }
             return TRUE;     
           }
     }
     pstate->pos.x=best_x;
     pstate->pos.y=best_y;
     if (side == S_RIGHT) {
	  pstate->dir += ((pstate->dir > 0) ? -180.0 : 180.0) ;
	  pstate->pos.x *= -1.0 ;
	  pstate->pos.y *= -1.0 ;
     }
     return TRUE;     
}
