/*
 *	Header:
 *	File: netif.c
 *	Author: Noda Itsuki
 *      Modified by : Ehsan Foroughi ( Sharif University of Tech. )
 *	Date: 2000/08/1
 *	EndHeader:
 */

#include <stdio.h>
#include <stdlib.h>

#include "libsclient.h"

const FlagInfo Flags[] = FLAGINFOTAB ;

Bool receive_info(Socket* socket, char* buffer, int size)
{
  int ret = FALSE ;
  char save = '\0';

  while(TRUE) {
    if (receive_message(socket, buffer, size)) {
      ret = TRUE ;
      save = buffer[0] ;
    }
    else
      break ;
  }

  buffer[0] = save ;
  return ret ;
}

Bool scan_info(char *buffer, ReceivedInfo *sinfo, char *tname)
{
  Bool scan_visual_info(char *buffer, SeeInfo *sinfo, char *teamname) ;
  Bool scan_auditorial_info(char *buffer, HearInfo *hinfo) ;
  Bool scan_senseBody_info(char *buffer, SenseBodyInfo *sbinfo) ;
  char com[COMMAX] ;
  char teamname[NAMEMAX];
  sprintf(teamname,"\"%s\"",tname);

  sscanf(buffer, "(%s ", com) ;

  if (!strcmp(com, "see")) {
    sinfo->infotype = SensorT;
    sinfo->info.sensor.type = SIT_See ;
    return scan_visual_info(buffer, &(sinfo->info.sensor.info.see), teamname) ;
  }
  else if (!strcmp(com, "hear")) {
    sinfo->infotype = SensorT;
    sinfo->info.sensor.type = SIT_Hear ;
    return scan_auditorial_info(buffer, &(sinfo->info.sensor.info.hear)) ;
  }
  else if (!strcmp(com, "sense_body")) {
    sinfo->infotype = SensorT;
    sinfo->info.sensor.type = SIT_SenseBody ;
    return scan_senseBody_info(buffer, &(sinfo->info.sensor.info.senseBody)) ;
  }
  else if (!strcmp(com, "server_param")) {
    sinfo->infotype = ParamT;
    sinfo->info.param.type = PIT_ServerParam ;
    return strcpy(sinfo->info.param.info.server_param,buffer) ;
  }
  else if (!strcmp(com, "player_param")) {
    sinfo->infotype = ParamT;
    sinfo->info.param.type = PIT_PlayerParam ;
    return strcpy(sinfo->info.param.info.player_param,buffer) ;
  }
  else if (!strcmp(com, "player_type")) {
    sinfo->infotype = ParamT;
    sinfo->info.param.type = PIT_PlayerType ;
    return strcpy(sinfo->info.param.info.player_type,buffer) ;
  }
  else if (!strcmp(com, "fullstate")) {
    sinfo->infotype = FullStateT;
    return strcpy(sinfo->info.fullstate,buffer) ;
  }
  else if (!strcmp(com, "see_global")) {
    sinfo->infotype = SeeGlobalT;
    return strcpy(sinfo->info.see_global,buffer) ;
  }
  else if (!strcmp(com, "error")) {
    sinfo->infotype = ErrorT;
    return strcpy(sinfo->info.error_msg,buffer) ;
  }
  else {
    return FALSE ;
  }

  return TRUE ;
}


char *next_token(char *buf)
{
  int i = 0 , j = 0;

  while (TRUE) {
    if (*buf == '(')
      i++ ;
    else if (*buf == ')') {
      if (i == 0) {
	*buf = '\0' ;
	return NULL ;
      }
      i-- ;
    } else if (*buf == '"') {
      j=1-j;
    } else if ((*buf == ' ') && (i == 0) && (j == 0))
      break ;
    else if (*buf == '\0')
      return NULL ;
    buf++ ;
  }

  *buf = '\0' ;
  buf++ ;

  return buf ;
}


Bool scan_visual_info(char *buffer, SeeInfo *sinfo, char *teamname)
{
  char *next, *obj, *param, *tmp ;
  int num = 0 ;
  double p[6];
  int num_param = 0;
  int j = 0;
  int outVCP = -1;

  sscanf(buffer, "(see %d ", &(sinfo->time)) ;
  buffer = next_token(buffer + 1) ;
  if(buffer != NULL) buffer = next_token(buffer) ;

  while(buffer != NULL) {
    next = next_token(buffer++) ;
    param = next_token(buffer++) ;
  
    num_param = 0;
    while (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      p[num_param] = atof(tmp);
      num_param++;
    }

    if (num_param == 1) {
      sinfo->obj[num].dist = DIST_ERR;
      sinfo->obj[num].dir = p[0] + ((p[0] >= 0.0) ? 0.5 : -0.5);
      sinfo->obj[num].dist_chng = CHNG_ERR;
      sinfo->obj[num].dir_chng = CHNG_ERR;
      sinfo->obj[num].body_dir = DIR_ERR;
      sinfo->obj[num].face_dir = DIR_ERR;
    } else if (num_param == 2) {
      sinfo->obj[num].dist = p[0];
      sinfo->obj[num].dir = p[1] + ((p[1] >= 0.0) ? 0.5 : -0.5);
      sinfo->obj[num].dist_chng = CHNG_ERR;
      sinfo->obj[num].dir_chng = CHNG_ERR;
      sinfo->obj[num].body_dir = DIR_ERR;
      sinfo->obj[num].face_dir = DIR_ERR;
    } else if (num_param == 4) {
      sinfo->obj[num].dist = p[0];
      sinfo->obj[num].dir = p[1] + ((p[1] >= 0.0) ? 0.5 : -0.5);
      sinfo->obj[num].dist_chng = p[2];
      sinfo->obj[num].dir_chng = p[3];
      sinfo->obj[num].body_dir = DIR_ERR;
      sinfo->obj[num].face_dir = DIR_ERR;
    } else if (num_param == 6) {
      sinfo->obj[num].dist = p[0];
      sinfo->obj[num].dir = p[1] + ((p[1] >= 0.0) ? 0.5 : -0.5);
      sinfo->obj[num].dist_chng = p[2];
      sinfo->obj[num].dir_chng = p[3];
      sinfo->obj[num].body_dir = p[4] + ((p[4] >= 0.0) ? 0.5 : -0.5);
      sinfo->obj[num].face_dir = p[5] + ((p[5] >= 0.0) ? 0.5 : -0.5);
    } else {
      printf("scan_visual_info: parse error\n") ;
      return FALSE;
    }

    obj = next_token(buffer) ;

    if (*buffer=='b') { 
      sinfo->obj[num].type = OT_Ball ;
      sinfo->obj[num].inViewConeP = TRUE ;
    }
    else if(*buffer=='B') {
      sinfo->obj[num].type = OT_Ball ;
      sinfo->obj[num].inViewConeP = FALSE ;
    }
    else if(*buffer=='p') {
      sinfo->obj[num].type = OT_Player ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.player.team = WT_Unknown ;
      sinfo->obj[num].id.player.unum = UNUM_ERR ;
      if (obj != NULL) {
	buffer = obj ;
	obj = next_token(buffer) ;
	if (!strcmp(buffer, teamname))
	  sinfo->obj[num].id.player.team = WT_Our ;
	else
	  sinfo->obj[num].id.player.team = WT_Opponent ;
				
	if (obj != NULL) {
	  (void)next_token(obj) ;
	  sinfo->obj[num].id.player.unum = atoi(obj) ;
	}
      }
    }
    else if(*buffer=='P') {
      sinfo->obj[num].type = OT_Player ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.player.team = WT_Unknown ;
      sinfo->obj[num].id.player.unum = UNUM_ERR ;
    }
    else if(*buffer=='g') {
      sinfo->obj[num].type = OT_Goal ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.side = S_UNKNOWN ;
      if (obj != NULL) {
        buffer = obj;
	if (*buffer == 'l')
	  sinfo->obj[num].id.side = S_LEFT ;
	else if (*buffer == 'r')
	  sinfo->obj[num].id.side = S_RIGHT ;
      }
    }
    else if(*buffer=='G') {
      sinfo->obj[num].type = OT_Goal ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.side = S_UNKNOWN ;
    }
    else if(*buffer=='f') {
      sinfo->obj[num].type = OT_Flag ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.fpos = FP_Unknown ;
      if (obj != NULL) {
	int namelen ;
	int i ;
	buffer = obj ;
	for(namelen = 0  ; 
	    (buffer[namelen] != ')')&&(buffer[namelen] != '\000') ;
	    namelen++) ;
	for(i = j ; i < FP__MAX ; i++) {
	  if(namelen == strlen(Flags[i].namestr) &&
	     !strncmp(buffer,Flags[i].namestr,namelen)) {
	    sinfo->obj[num].id.fpos = Flags[i].name ;
            j = i + 1;
	    if (outVCP >= 0) {
	      sinfo->obj[outVCP].end = j - 2;
	      if (sinfo->obj[outVCP].begin == sinfo->obj[outVCP].end) {
		sinfo->obj[outVCP].inViewConeP = TRUE;
		sinfo->obj[num].id.fpos = (FlagPos)sinfo->obj[outVCP].begin;
	      }
	      outVCP = -1;
	    }
	    break ;
	  }
	}
      }
    }
    else if(*buffer=='F') {
      sinfo->obj[num].type = OT_Flag ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.fpos = FP_Unknown ;
      sinfo->obj[num].begin = j;
      sinfo->obj[num].end = FP_r_b_30;
      outVCP = num;
    }
    else if(*buffer=='l') {
      sinfo->obj[num].type = OT_Line ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.lpos = LP_Unknown ;
      if (obj != NULL) {
	buffer = obj ;
	if (*buffer == 'l')
	  sinfo->obj[num].id.lpos = LP_L ;
	else if (*buffer == 't')
	  sinfo->obj[num].id.lpos = LP_T ;
	else if (*buffer == 'b')
	  sinfo->obj[num].id.lpos = LP_B ;
	else if (*buffer == 'r')
	  sinfo->obj[num].id.lpos = LP_R ;
      }
    }
    else if(*buffer=='L') {
      sinfo->obj[num].type = OT_Line ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.lpos = LP_Unknown ;
    }
    else {
      printf("scan_visual_info: miss obj\n") ;
      return FALSE ;
    }

    buffer = next ;
    num++ ;
  }
  sinfo->n_obj = num ;

  return TRUE ;
}


Bool scan_auditorial_info(char *buffer, HearInfo *hinfo)
{
  char tmp[64] ;
  int i ;

  sscanf(buffer, "(hear %d %s", &(hinfo->time), tmp) ;
  buffer = next_token(buffer + 1) ;
  buffer = next_token(buffer) ;
  buffer = next_token(buffer) ;

  for(i = strlen(buffer); i >= 0; i--)
    if (buffer[i] == ')') {
      buffer[i] = '\0' ;
      break ;
    }

  strcpy(hinfo->message, buffer) ;

  if (!strcmp(tmp, "referee")) {
    hinfo->sender = ST_Referee ;
    hinfo->direction = 0.0 ;
  }
  else if (!strcmp(tmp, "self")) {
    hinfo->sender = ST_Self ;
    hinfo->direction = 0.0 ;
  }
  else if (!strcmp(tmp, "online_coach_left")) {
    hinfo->sender = ST_OLCoach_l ;
    hinfo->direction = 0.0 ;
  }
  else if (!strcmp(tmp, "online_coach_right")) {
    hinfo->sender = ST_OLCoach_r ;
    hinfo->direction = 0.0 ;
  }
  else {
    hinfo->sender = ST_Player ;
    hinfo->direction = atof(tmp) ;
  }

  return TRUE ;
}

Bool scan_senseBody_info(char *buffer, SenseBodyInfo *sbinfo) {
     char *next, *obj, *param, *tmp ;

     sscanf(buffer, "(sense_body %d ", &(sbinfo->time)) ;
     buffer = next_token(buffer + 1) ;
     buffer = next_token(buffer) ;

     sbinfo->countKick = -1 ;      sbinfo->countDash = -1 ; 
     sbinfo->countTurn = -1 ;      sbinfo->countSay = -1 ; 
     sbinfo->countTurnNeck = -1;   sbinfo->countCatch = -1 ; 
     sbinfo->countMove = -1;       sbinfo->countChangeView = -1 ; 

     while(buffer != NULL) {
	  next = next_token(buffer) ;
	  buffer++ ;
	  param = next_token(buffer) ;
	  
	  if (!strcmp(buffer,"view_mode")) {
	       char *nparam ;
	       nparam = next_token(param) ;
	       if(!strcmp(param,"high")) {
		    sbinfo->viewQuality = VQ_high ;
	       } else {
		    sbinfo->viewQuality = VQ_low ;
	       }
	       if(!strcmp(nparam,"wide)")) {
		    sbinfo->viewWidth = VW_Wide ;
	       } else if(!strcmp(nparam,"normal)")) {
		    sbinfo->viewWidth = VW_Normal ;
	       } else {
		    sbinfo->viewWidth = VW_Narrow ;
	       }
	  } else if (!strcmp(buffer,"stamina")) {
	       sscanf(param,"%lf %lf)",&sbinfo->stamina, &sbinfo->effort) ;
	  } else if (!strcmp(buffer,"speed")) {
	       sscanf(param,"%lf %lf",&sbinfo->speed_val,&sbinfo->speed_dir);
	  } else if (!strcmp(buffer,"head_angle")) {
	       sscanf(param,"%lf",&sbinfo->head_angle) ;
	  } else if (!strcmp(buffer,"kick")) {
	       sscanf(param,"%d",&sbinfo->countKick) ;
	  } else if (!strcmp(buffer,"dash")) {
	       sscanf(param,"%d",&sbinfo->countDash) ;
	  } else if (!strcmp(buffer,"turn")) {
	       sscanf(param,"%d",&sbinfo->countTurn) ;
	  } else if (!strcmp(buffer,"say")) {
	       sscanf(param,"%d",&sbinfo->countSay) ;
	  } else if (!strcmp(buffer,"turn_neck")) {
	       sscanf(param,"%d",&sbinfo->countTurnNeck) ;
	  } else if (!strcmp(buffer,"catch")) {
	       sscanf(param,"%d",&sbinfo->countCatch) ;
	  } else if (!strcmp(buffer,"move")) {
	       sscanf(param,"%d",&sbinfo->countMove) ;
	  } else if (!strcmp(buffer,"change_view")) {
	       sscanf(param,"%d",&sbinfo->countChangeView) ;
	  } else {
	       printf("Illegal sense_body info:(%s %s\n",
		       buffer,param);
	       return FALSE ;
	  } 
	  buffer = next ;
     }
     sbinfo->head_angle += (sbinfo->head_angle >= 0.0) ? 0.5 : -0.5;

     return TRUE ;
}

Bool scan_info_online_coach(char *buffer, ReceivedInfo *sinfo, char *tname)
{
  Bool scan_visual_info_online_coach(char *buffer, SeeInfo *sinfo, char *teamname) ;
  Bool scan_auditorial_info(char *buffer, HearInfo *hinfo) ;
  char com[COMMAX] ;
  char teamname[NAMEMAX];
  sprintf(teamname,"\"%s\"",tname);

  sscanf(buffer, "(%s ", com) ;

  if (!strcmp(com, "see")) {
    sinfo->infotype = SensorT;
    sinfo->info.sensor.type = SIT_See ;
    return scan_visual_info_online_coach(buffer, &(sinfo->info.sensor.info.see), teamname) ;
  }
  else if (!strcmp(com, "hear")) {
    sinfo->infotype = SensorT;
    sinfo->info.sensor.type = SIT_Hear ;
    return scan_auditorial_info(buffer, &(sinfo->info.sensor.info.hear)) ;
  }
  else {
    return FALSE ;
  }

  return TRUE ;
}


Bool scan_visual_info_online_coach(char *buffer, SeeInfo *sinfo, char *teamname)
{
  char *next, *obj, *param, *tmp ;
  int num = 0 ;

printf("%s\n", buffer);

  sscanf(buffer, "(see %d ", &(sinfo->time)) ;
  buffer = next_token(buffer + 1) ;
  if(buffer != NULL) buffer = next_token(buffer) ;

  while(buffer != NULL) {
    next = next_token(buffer++) ;
    param = next_token(buffer++) ;

    if (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      sinfo->obj[num].dist = atof(tmp);
    }
    else
      sinfo->obj[num].dist = DIST_ERR ;

    if (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      sinfo->obj[num].dir = atof(tmp) ;
    }
    else
      sinfo->obj[num].dir = DIR_ERR ;

    if (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      sinfo->obj[num].dist_chng = atof(tmp) ;
    }
    else
      sinfo->obj[num].dist_chng = CHNG_ERR ;

    if (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      sinfo->obj[num].dir_chng = atof(tmp) ;
    }
    else
      sinfo->obj[num].dir_chng = CHNG_ERR ;

    if (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      sinfo->obj[num].face_dir = atof(tmp) ;
    }
    else
      sinfo->obj[num].face_dir = DIR_ERR ;

    obj = next_token(buffer) ;

    if (*buffer=='b') { 
      sinfo->obj[num].type = OT_Ball ;
      sinfo->obj[num].inViewConeP = TRUE ;
    }
    else if(*buffer=='B') {
      sinfo->obj[num].type = OT_Ball ;
      sinfo->obj[num].inViewConeP = FALSE ;
    }
    else if(*buffer=='p') {
      sinfo->obj[num].type = OT_Player ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.player.team = WT_Unknown ;
      sinfo->obj[num].id.player.unum = UNUM_ERR ;
      if (obj != NULL) {
	buffer = obj ;
	obj = next_token(buffer) ;
	if (!strcmp(buffer, teamname))
	  sinfo->obj[num].id.player.team = WT_Our ;
	else
	  sinfo->obj[num].id.player.team = WT_Opponent ;
				
	if (obj != NULL) {
	  (void)next_token(obj) ;
	  sinfo->obj[num].id.player.unum = atoi(obj) ;
	}
      }
    }
    else if(*buffer=='P') {
      sinfo->obj[num].type = OT_Player ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.player.team = WT_Unknown ;
      sinfo->obj[num].id.player.unum = UNUM_ERR ;
    }
    else if(*buffer=='g') {
      sinfo->obj[num].type = OT_Goal ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.side = S_UNKNOWN ;
      if (obj != NULL) {
	buffer = obj ;
	obj = next_token(buffer) ;
	if (*buffer == 'l')
	  sinfo->obj[num].id.side = S_LEFT ;
	else if (*buffer == 'r')
	  sinfo->obj[num].id.side = S_RIGHT ;
      }
    }
    else if(*buffer=='G') {
      sinfo->obj[num].type = OT_Goal ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.side = S_UNKNOWN ;
    }
    else if(*buffer=='f') {
      sinfo->obj[num].type = OT_Flag ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.fpos = FP_Unknown ;
      if (obj != NULL) {
	int namelen ;
	int i ;
	buffer = obj ;
	for(namelen = 0  ; 
	    (buffer[namelen] != ')')&&(buffer[namelen] != '\000') ; 
	    namelen++) ;
	for(i = 0 ; i < FP__MAX ; i++) {
	  if(namelen == strlen(Flags[i].namestr) &&
	     !strncmp(buffer,Flags[i].namestr,namelen)) {
	    sinfo->obj[num].id.fpos = Flags[i].name ;
	    break ;
	  }
	}
      }
    }
    else if(*buffer=='F') {
      sinfo->obj[num].type = OT_Flag ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.fpos = FP_Unknown ;
    }
    else if(*buffer=='l') {
      sinfo->obj[num].type = OT_Line ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.lpos = LP_Unknown ;
      if (obj != NULL) {
	buffer = obj ;
	obj = next_token(buffer) ;
	if (*buffer == 'l')
	  sinfo->obj[num].id.lpos = LP_L ;
	else if (*buffer == 't')
	  sinfo->obj[num].id.lpos = LP_T ;
	else if (*buffer == 'b')
	  sinfo->obj[num].id.lpos = LP_B ;
	else if (*buffer == 'r')
	  sinfo->obj[num].id.lpos = LP_R ;
      }
    }
    else if(*buffer=='L') {
      sinfo->obj[num].type = OT_Line ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.lpos = LP_Unknown ;
    }
    else {
      printf("scan_visual_info: miss obj\n") ;
      return FALSE ;
    }

    buffer = next ;
    num++ ;
  }
  sinfo->n_obj = num ;

  return TRUE ;
}
