/*
  Player header file : 
    settings and defauult values
    program constants
    type definitions

  Ehsan Foroughi,     Mehran Motamed,
  Pooya Karimian,     Hamed  Hatami, 
           M.Hossein Fardad.

  ParseAI team,
  Sharif University
  Tehran, Iran

*/

#ifndef _PLAYER_H
#define _PLAYER_H

#ifndef _player_extern_
#define _player_extern_ extern
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "libsclient-7.00/include/libsclient.h"

// true if you want to log all things into teamname_unum.log , false if time is critical
#define LOGIT              true             //(player.unum==11)

// Sould it send data to a coach monitor? e.g. (player.unum==7)
#define MONITOR_ENABLED    false

#define DEBUG_MODE         false            // Debugging state
//#define DEBUG_ON

// Sould Log these parts of information
#define LOG_RCV            true
#define LOG_SEND           true
#define LOG_PLAYER_ST      true
#define LOG_SKILL          true
#define LOG_BEHAVIOR       true
#define LOG_ALL_WORLD_ST   true
#define LOG_FULLSTATE      false

#define REPORT_ERR         true

#define DEFAULTSERVER      "localhost"
#define DEFAULTPLAYERPORT  6000
#define DEFAULTCOACHPORT   6002

#define MONITOR_ADDR       "localhost"
#define MONITOR_PORT       7000

// - Some of server options
#define SERVER_VERSION     7.00
#define GAME_CYCLES        6000

#define mytypes           types[player.idtype]

#define FIELD_LENGTH       PITCH_LENGTH
#define FIELD_WIDTH        PITCH_WIDTH
//#define GOAL_WIDTH         (params.server.gwidth)
#define STAMINA_MAX        (params.server.stamina_max)
#define STAMINA_INC_MAX    (params.server.stamina_inc)
#define PLAYER_SIZE        (params.mytypes.player_size)
#define PLAYER_DECAY       (params.mytypes.player_decay)
#define DASH_POWER_RATE    (params.mytypes.dash_power_rate)
#define MAX_DASH_POWER     (params.server.maxp)
#define MIN_DASH_POWER     (params.server.minp)
#define INERTIA_MOMENT     (params.mytypes.inertia_moment)
#define BALL_SIZE          (params.server.bsize)
#define BALL_DECAY         (params.server.bdecay)
#define MAX_KICK_POWER     (params.server.maxp)
#define KICKABLE_MARGIN    (params.mytypes.kickable_margin)
#define KICKABLE_AREA      (KICKABLE_MARGIN+PLAYER_SIZE+BALL_SIZE)-0.1
#define KICK_POWER_RATE    (params.server.kprate)
#define VISIBLE_DISTANCE   (params.server.visdist)

#define CATCHABLE_AREA_L   2.0
#define CATCHABLE_AREA_W   1.0
#define CATCH_BAN_CYCLE    5
#define COLLISION_VELOCITY -0.1
#define VIEW_ANGLE         45

#define CONE_ANGLE         45

// - Progrom default values
#define GOALER_UNUM        1
#define SEE_CORRECTNESS    95.0             // reliability of sight
#define CORRECTNESS_DIST_W 0.25
#define CORRECTNESS_DECAY  8.0              // decay of reliability
#define HEAR_CRR_DECAY     5.0
#define MIN_USEFUL_CRR     25.0
#define RELIEABLE_CRR      50.0
#define LEAST_FIN_BALL_VEL 0.05

#define GOAL_LOW_Y         -(GOAL_WIDTH/2-GOAL_CARE_RANGE)
#define GOAL_HIGH_Y        (GOAL_WIDTH/2-GOAL_CARE_RANGE)
#define GOAL_X             (FIELD_LENGTH/2)

#define GHOST_CORRECTNESS  20.0
#define GHOST_RANGE        18.0
#define GHOST_CARE_ANGLE   10.0

#define MAX_UNUM_FIND_DIST 1.2

#define GOOD_POSITIONING   true

// - Range of values and values of common error in data
#define FLOAT_ERROR        0.001            // common floating error
#define LOC_ERROR          1                // location error in meters
#define DIR_ERROR          2                // direction error in degrees
#define MAX_PREDICT_TURNS  20

#define GOTO_LOC_DIR_ERR   20.0 // - *
#define FAR_DIST_2DEST     1.2  // *
#define NEAR_ENOUGH_2DEST  0.6  // *
#define LOW_PLAYER_SPEED   0.2
#define GOAL_CARE_RANGE    2  // - *
#define GOALER_STAND_RAD   8.0
#define OFFSIDE_CARE_RANGE 2.5

#define BALL_CONTROL_RANGE 3.0
#define BALL_CONTROL_TURNS 3
#define BALL_CONTROL_SPEED 0.7

#define DOUBLE_KICK_SPEED  0.3
#define DOUBLE_KICK_RANGE  (KICKABLE_AREA/2)

#define POS_VAL_RANGE_RAD  15
#define MAX_PASS_RADIUS    40
#define MIN_PASS_RADIUS    5
#define GOALER_PASS_FACTOR 0.0

#define DRIBBLE_DANGER_RADIUS 2.2

// - Strategy predefined values
#define INTERPOLATION      3

#define say_time()         ((((((long)world.time.clock*167)%5)+1)==player.unum%5)&&(world.time_running))
#define Time2Say()         (say_time()||(sensor.comm.sth2say))
//#define Time2Say()         ((((((long)world.time.clock*167)%11)+1)==player.unum)&&(world.time_running))
//#define Time2Say()         true
#define SENT_COMMANDS_MEM  4096
#define timing_index(x)    (x.clock+x.stopped)%SENT_COMMANDS_MEM
#define sent_command(x)    sent_commands[timing_index(x)]
#define just_sent_command  sent_command(world.time)
#define last_sent_command  sent_command(world.lasttime)

// - Main Program Constants
#define MAIN_BUFFER_SIZE   4096             // receive buffer size
#define MAX_QUEUE_MESSAGE  16

#define MAX_SAY_SIZE       512

#define MAX_MONITOR_SIZE   1024

#define AUTO_TERMINATE_SEC 5

#define TEAM_SIGNATURE     "PAI"

#define min(a,b)  ((a) > (b) ? (b) : (a))
#define max(a,b)  ((a) < (b) ? (b) : (a))

#define timeval(a) (a.clock+a.stopped)

#endif /* !_PLAYER_H */
