#define _skills_move_extern_
#include "move.h"


int c_move_utils::least_turns_needed_2move(double length)
{
  if (!calculated) calculate();
  if (length>FIELD_LENGTH*2.0) length=FIELD_LENGTH*2.0;
  return(mlar[(int)(length*2.0)]);
}

void c_move_utils::calculate(void)
{
  double l, d, v;
  int t, ti;
  
  for (l=0;l<FIELD_LENGTH*2.0-1;l+=0.5) {
    v = 0.0;
    d = 0.0;
    t = 0;
    while (d<l) {
      d += v;
      v *= PLAYER_DECAY;
      v += MAX_DASH_POWER*DASH_POWER_RATE;
      t++;
    };  
    ti = (int)(l*2.0);
    assert(ti<(int)(FIELD_LENGTH)*4);
    mlar[ti] = t;
  };

  calculated = true;
}

int c_move_skills::who_controls_ball(void)
{
  double x, y, iv;
  int i, tr, btr, bp;
  
  bp = player.unum;
  btr = 1000;

  for (i=1;i<=N_PLAYER_PER_TEAM;i++) {
    x = world.tmate[i].x;
    y = world.tmate[i].y;
    iv = speed(world.tmate[i]);
    tr = 0;
    while ((least_turns_needed_2move(dist(x-estimate_ball_x(tr),y-estimate_ball_y(tr)),iv,stamina_usage(5,world.tmate[i].stamina))>tr)&&(tr<btr))
      tr++;
    if (tr<btr) {
      btr = tr;
      bp = i;
    }
  }

  for (i=1;i<=N_PLAYER_PER_TEAM;i++) {
    x = world.opp[i].x;
    y = world.opp[i].y;
    iv = speed(world.opp[i]);
    tr = 0;
    while ((least_turns_needed_2move(dist(x-estimate_ball_x(tr),y-estimate_ball_y(tr)),iv,90)>tr)&&(tr<btr))
      tr++;
    if (tr<btr) {
      btr = tr;
      bp = -i;
    }
  }

  return bp;
}

int c_move_skills::can_get2ball_fastest(void)
{
  double x, y, iv;
  int i, tr, btr, bp;
  
  bp = player.unum;
  btr = 100;

  for (i=1;i<=N_PLAYER_PER_TEAM;i++) {
    x = world.tmate[i].x;
    y = world.tmate[i].y;
    iv = speed(world.tmate[i]);
    tr = 0;
    while ((least_turns_needed_2move(dist(x-estimate_ball_x(tr),y-estimate_ball_y(tr)),iv,stamina_usage(5,world.tmate[i].stamina))>tr)&&(tr<btr))
      tr++;
    if (tr<btr) {
      btr = tr;
      bp = i;
    }
  }

  return bp;
}

int c_move_skills::least_turns_needed_2move(double length, double iv, double dashpw)
{
  double l, d = 0, v = iv;
  int t = 0;
  
  l = length;
  while (d<l) {
    d += v;
    v *= PLAYER_DECAY;
    v += dashpw*DASH_POWER_RATE;
    t++;
  };  
  return(t);
}


int c_move_skills::turns_needed_2get2loc(double x, double y, double maxdash)
{
  int t = least_turns_needed_2move(dist_2me(x,y),speed(world.body.loc),maxdash);
  if (fabs(reldir_2me(x,y))>GOTO_LOC_DIR_ERR) t++;
  return(t);
}

int c_move_skills::ballturns_2get_mostnear(double maxdash)
{
  int tr = 0;

  while (turns_needed_2get2loc(estimate_ball_x(tr),estimate_ball_y(tr),maxdash)>tr)
    tr++;

  return tr;
}

// 0 (Zero) if not blocked + if tmate - if opp
int c_move_skills::way_blocked(double x, double y)
{
  int i;
  double d, r;
  double dir = dir_from_me(x,y);

  if (dist_2me(world.ball)<BLOCKING_BALL_DIST) return 0;

  for(i=1;i<=N_PLAYER_PER_TEAM;i++) 
    if (world.opp[i].correctness>MIN_USEFUL_CRR) {
      d = normalize_angle(dir_from_me(world.opp[i])-dir);
      r = dist_2me(world.opp[i]);
#ifdef DEBUG_ON
      sprintf(debugbuf,"%d %.2f %.2f",-i,d,r);
      logmsg(debugbuf);
#endif
      if ((fabs(d)<100) && (r<BLOCKING_DIST))
        if (fabs(sin(Deg2Rad(d))*r)<KICKABLE_AREA)
          return -i;
    };

  if (TEAM_MATES_BLOCK) 
    for(i=1;i<=N_PLAYER_PER_TEAM;i++) 
      if ((world.tmate[i].correctness>MIN_USEFUL_CRR)&&(i!=player.unum)) {
        d = normalize_angle(dir_from_me(world.tmate[i])-dir);
        r = dist_2me(world.tmate[i]);
#ifdef DEBUG_ON
        sprintf(debugbuf,"%d %.2f %.2f",i,d,r);
        logmsg(debugbuf);
#endif
        if ((fabs(d)<90) && (r<BLOCKING_DIST))
          if (fabs(sin(Deg2Rad(d))*r)<KICKABLE_AREA)
            return i;
      };

  return 0;
}


void c_move_skills::move_2dest(Command &com, double x, double y, double maxdash)
{
  double dashpower;
  double dir;

  blocking_unum = way_blocked(x,y);

  if (blocking_unum==0) is_passing_block = false;

  if (blocking_unum!=0) {
    if (is_passing_block) {
      com_dash(com,maxdash);
      is_passing_block = false;
    } else {
      if (blocking_unum>0) dir = (reldir_2me(world.tmate[blocking_unum])>0)?-PASS_BLOCK_ANGLE:PASS_BLOCK_ANGLE;
      else dir = (reldir_2me(world.opp[blocking_unum])>0)?-PASS_BLOCK_ANGLE:PASS_BLOCK_ANGLE;
      com_turn(com,normalize_angle(reldir_2me(x,y)+dir));
      is_passing_block = true;
    };
  } else if (fabs(reldir_2me(x,y))>GOTO_LOC_DIR_ERR)
    com_turn(com,reldir_2me(x,y));
  else if (dist_2me(x,y)>FAR_DIST_2DEST)
    com_dash(com,maxdash);
  else if (dist_2me(x,y)<NEAR_ENOUGH_2DEST) {
    if (speed(world.body.loc)<LOW_PLAYER_SPEED)
      com_turn(com,reldir_2me(world.ball));
    else {
      dashpower = -(speed(world.body.loc))/DASH_POWER_RATE;
      if (dashpower<MIN_DASH_POWER) dashpower=MIN_DASH_POWER;
      com_dash(com,dashpower);
    }
  } else {
    if (world.time.clock-last_dribble_time<5) com_dash(com,maxdash);
    else com_dash(com,maxdash/2);
  }
}

void c_get_2ball::update_bytime(void)
{
  com2send.type = NO_COMMAND;

  if ((client.role==GOALER)||(world.time.clock-last_dribble_time<5)) maxdash = stamina_usage(6);
  else maxdash = stamina_usage(5);
}

bool c_get_2ball::possible(void)
{
  return true;
}

double c_get_2ball::weight(void)
{
  double val = 100 - 2*dist_2me(world.ball);
  if (val<0) val=0;
  return val;
}

void c_get_2ball::perform(void)
{
  calculate();
  
  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file, "\nLogging Skills state in time %d :", world.time.clock);
    fprintf(client.log.file, "\n  Skill : Getting to ball skill acting.");
    if (is_passing_block) 
      fprintf(client.log.file, "\n    Player with unum %d blocking way : Tring to move beside it.", blocking_unum);
    fprintf(client.log.file, "\n    Turns : %d , Destination : (%.2f,%.2f)\n", turns, estimate_ball_x(turns), estimate_ball_y(turns));
  };
  perform_command(com2send);
}

int c_get_2ball::needed_turns(void)
{
  calculate();
  return turns;
}

void c_get_2ball::calculate(void)
{
  double destx, desty, bd;
  int nmo;
  
//  turns = ballturns_2get_mostnear(maxdash);
  turns = ballturns_2get_mostnear(stamina_usage(5,world.body.stamina));
  destx = estimate_ball_x(turns);
  desty = estimate_ball_y(turns);

  nmo = nearmost_player(world.body.loc, OPPONENTS);
  bd = dist_2me(world.ball);

  if ((bd<3)&&(world.body.loc.x<world.ball.x))
    if ((dist2point(world.ball.x,world.ball.y,world.opp[nmo])<KICKABLE_AREA)&&(world.opp[nmo].correctness>MIN_USEFUL_CRR))
      destx = world.opp[nmo].x-0.4;

  if ((bd<3)&&(world.body.loc.x>world.ball.x))
    if ((world.opp[nmo].correctness>MIN_USEFUL_CRR)&&
        ((world.body.loc.x>world.opp[nmo].x)&&(world.opp[nmo].x>world.ball.x))&&
	((world.body.loc.y-world.opp[nmo].y)*(world.opp[nmo].y-world.ball.y)>0))
      destx = world.opp[nmo].x-2;


  move_2dest(com2send, destx, desty, maxdash);
}

void c_goto_playing_post::update_bytime(void)
{
  com2send.type = NO_COMMAND;
}

bool c_goto_playing_post::possible(void)
{
  return true;
}

double c_goto_playing_post::weight(void)
{
  double val = 10+2*dist_2me(world.ball);
  if (val>100) val=100;
  if (can_kickball()) val = 10;
  return val;
}

void c_goto_playing_post::perform(void)
{
  calculate();
  
  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file, "\nLogging Skills state in time %d :", world.time.clock);
    fprintf(client.log.file, "\n  Skill : Goto playing post skill acting.");
    fprintf(client.log.file, "\n    stand position (x,y) : (%.2f,%.2f)\n",destx,desty);
  };
  perform_command(com2send);
}

double c_goto_playing_post::position_in_field_value(double x, double y)
{
  double return_value;
  double d;
  double rx=36,ry=25;
  
  d = dist(x/rx,y/ry);
  return_value=100-fabs(1-d)*40;

  if (return_value>100) return_value=100;
  if (return_value<0) return_value=0;
  return return_value;
}

double c_goto_playing_post::standing_position_value(double x, double y)
{
  double d;

  d = dist(stand_pos_x()-x,stand_pos_y()-y) - STAND_RANGE;
  if (d<0) d = 0;
  if (d>15) d = 15;
  return (100-100*d/15);
}

double c_goto_playing_post::opponents_value(double x, double y)
{
  double d=0, w=0;
  int nmo;
  double dir, dox, doy;


  switch (client.role) {
    case GOALER:
      w = 50;
    break;
    case DEFEND:
      nmo = nearmost_player(x, y, OPPONENTS);
      dir = direction(world.ball.x-world.opp[nmo].x,world.ball.y-world.opp[nmo].y);
      dox = world.opp[nmo].x + cos(Deg2Rad(dir))*3;
      doy = world.opp[nmo].y + sin(Deg2Rad(dir))*3;
      d = dist(x-dox,y-doy);

      d -= NEAR_OPP_RANGE;
      if (d>16) d = 16;
      if (d<0) d = 0;
      w = 100 - sqrt(d)*25;
    break;
    case MIDDLE:
        d -= NEAR_OPP_RANGE;
        if (d>5) d = 5;
        if (d<0) d = 0;
        w = d*d + 25;

/*      if (who_controls_ball()>0) {
        d -= NEAR_OPP_RANGE;
        if (d>10) d = 10;
        if (d<0) d = 0;
        w = d*d;
      } else {
        d -= NEAR_OPP_RANGE;
        if (d>16) d = 16;
        if (d<0) d = 0;
        w = 100 - sqrt(d)*25;
      };
      w *= 0.5;
      w += 25;*/
    break;
    case ATTACK:
      nmo = nearmost_player(x, y, OPPONENTS);
      dir = direction(world.ball.x-world.opp[nmo].x,world.ball.y-world.opp[nmo].y);
      dox = world.opp[nmo].x - cos(Deg2Rad(dir))*3;
      doy = world.opp[nmo].y - sin(Deg2Rad(dir))*3;
      d = dist(x-dox,y-doy);

      d -= NEAR_OPP_RANGE;
      if (d>10) d = 10;
      if (d<0) d = 0;
      w = d*d;
      w = w/2 + 25;
    break;
  }

  return w;
}

double c_goto_playing_post::offside_value(double x, double y)
{
  double d = world.offside.opp.line - x + OFFSIDE_CARE_RANGE/2 - 1;
  
  if (d<0) d = 0;
  if (d>OFFSIDE_CARE_RANGE) d = OFFSIDE_CARE_RANGE;
  return (d/OFFSIDE_CARE_RANGE);
}

double c_goto_playing_post::land_value(double x, double y)
{
  double pos_field_val = position_in_field_value(x,y);
  double stand_pos_val = standing_position_value(x,y);
  double opponents_val = opponents_value(x,y);
  double offside_val = offside_value(x,y);

  double land_val = pos_field_val*0.1 + stand_pos_val*0.5 + opponents_val*0.4;
  land_val *= offside_val;  
  if (!in_region(x,y)) land_val *= 0.8;
  if ((kick_play_mode())&&(in_penalty_area(x,y))) land_val *= 0.2;

#ifdef DEBUG_ON
  sprintf(debugbuf, "Values x,y %.2f %.2f land %.2f pos %.2f stand %.2f opp %.2f", x, y, land_val, pos_field_val, stand_pos_val, opponents_val);
  logmsg(debugbuf);
#endif

  return land_val;
}

void c_goto_playing_post::calc_pos(void)
{
  double dir;
  int nmo = nearmost_player(destx, desty, OPPONENTS);

  switch (client.role) {
    case GOALER:
    break;
    case DEFEND:
      if (dist2point(destx,desty,world.opp[nmo])<12) {
        dir = direction(world.ball.x-world.opp[nmo].x,world.ball.y-world.opp[nmo].y);
	destx = world.opp[nmo].x + cos(Deg2Rad(dir))*2;
	desty = world.opp[nmo].y + sin(Deg2Rad(dir))*2;
      };
    break;
    case MIDDLE:
    break;
    case ATTACK:
    break;
  }
}

void c_goto_playing_post::calc_dest(void)
{
  double l, bl, i, j, dx, dy;

  destx = stand_pos_x();
  desty = stand_pos_y();

//  calc_pos();
//  return;

  dx = destx;
  dy = desty;

//  if (client.role == MIDDLE) return;

/*  if ((player.unum == 4)||(player.unum == 5)) {
    if ((dist_2me(destx,desty)<6)||(world.body.stamina<2000)) {
      destx = world.body.loc.x;
      desty = world.body.loc.y;
    }
    return;
  }*/

  if (client.role == GOALER) return;

  bl = -1;

  for(i=((dx-16>-FIELD_LENGTH)?(dx-16):-FIELD_LENGTH);i<=(dx+16);i+=1.5)
    for(j=((dy-16>-FIELD_WIDTH)?(dy-16):-FIELD_WIDTH);j<=(dy+16);j+=1.5) {
      l = land_value(i,j);
      if (l>bl) {
        bl = l;
	destx = i;
	desty = j;
      }
    }

  if (bl-land_value(world.body.loc.x,world.body.loc.y)<dist_2me(destx,desty)+4) {
    destx = world.body.loc.x;
    desty = world.body.loc.y;
  }

  if (!world.offside.inoff)
    if ((dist_2me(destx,desty)<2)||(world.body.stamina<1800)) {
      destx = world.body.loc.x;
      desty = world.body.loc.y;
    }

}

void c_goto_playing_post::calculate(void)
{
  calc_dest();
  
  if (dist_2me(destx,desty)>2.5)
    move_2dest(com2send, destx, desty, (client.role==GOALER)?stamina_usage(3):stamina_usage(2));
  else if (fabs(reldir_2me(world.ball))>30)
    com_turn(com2send,reldir_2me(world.ball));
}
