#define _world_command_extern_
#include "command.h"

void move(double x, double y)
{

  if(!send_com_move(&(client.sock),x,y)) logmsg("error sending move command.!!");
  else {
    if (LOGIT && LOG_SEND) {
      fprintf(client.log.file, "\nCommand send in time %d :", world.time.clock);
      fprintf(client.log.file, "\n  Move to location (%.2f,%.2f)\n", x, y);
    };
    world.body.loc.x = x;
    world.body.loc.y = y;
    world.body.loc.update_time = world.time;
//    world.body.loc.correctness -= CORRECTNESS_DECAY;
    if (world.body.ball_inhand) {
      world.ball.x = x;
      world.ball.y = y;
      world.ball.vx = 0.0;
      world.ball.vy = 0.0;
      world.ball.update_time = world.time;
//      world.ball.correctness -= CORRECTNESS_DECAY;
    };
  };
};

void turn(double moment)
{

  moment = effective_turn(moment);

  if(!send_com_turn(&(client.sock),moment)) logmsg("error sending turn command.!!");
  else {
    if (LOGIT && LOG_SEND) {
      fprintf(client.log.file, "\nCommand send in time %d :", world.time.clock);
      fprintf(client.log.file, "\n  Turn %.2f degrees\n", moment);
    };

    world.body.plcount.turn++;
    sensor.just_sent_command.type = TURN;
    sensor.just_sent_command.par1 = moment;
  };
};

void dash(double power)
{
  if(!send_com_dash(&(client.sock),power)) logmsg("error sending dash command.!!");
  else {
    if (LOGIT && LOG_SEND) {
      fprintf(client.log.file, "\nCommand send in time %d :", world.time.clock);
      fprintf(client.log.file, "\n  Dash %.2f\n", power);
    };

    world.body.plcount.dash++;
    sensor.just_sent_command.type = DASH;
    sensor.just_sent_command.par1 = power;
  };
};

void kick(double power, double dir)
{
  if(!send_com_kick(&(client.sock),power,dir)) logmsg("error sending kick command.!!");
  else {
    if (LOGIT && LOG_SEND) {
      fprintf(client.log.file, "\nCommand send in time %d :", world.time.clock);
      fprintf(client.log.file, "\n  Kick with power of %.2f in direction %.2f\n", power, dir);
    };

    world.body.plcount.kick++;
    sensor.just_sent_command.type = KICK;
    sensor.just_sent_command.par1 = power;
    sensor.just_sent_command.par2 = dir;
  };
};

void gcatch(double dir)
{
  if(!send_com_catch(&(client.sock),dir)) logmsg("error sending catch command.!!");
  else {
    if (LOGIT && LOG_SEND) {
      fprintf(client.log.file, "\nCommand send in time %d :", world.time.clock);
      fprintf(client.log.file, "\n  Catch in direction %.2f\n", dir);
    };
    world.body.plcount.gcatch++;
    world.body.last_catch = world.time;
//    world.body.ball_inhand = true;
  };
};

void turn_neck(double dir)
{
  if(!send_com_turn_neck(&(client.sock),dir)) logmsg("error sending turn_neck command.!!");
  else {
    if (LOGIT && LOG_SEND) {
      fprintf(client.log.file, "\nCommand send in time %d :", world.time.clock);
      fprintf(client.log.file, "\n  Turning neck %.2f degrees\n", dir);
    };
    world.body.plcount.turnneck++;
    sensor.just_sent_command.neck_moment = dir;
  };
};

void change_view(ViewWidth width,ViewQuality quality)
{
  if (width==world.view.width && quality==world.view.quality) 
    return; 
  if(!send_com_change_view(&(client.sock),width,quality)) logmsg("error sending change_view command.!!");
  else {
    if (LOGIT && LOG_SEND) {
      fprintf(client.log.file, "\nCommand send in time %d :", world.time.clock);
      fprintf(client.log.file, "\n  Changing View to width=%d quality=%d\n", (int)width,(int)quality);
    };
  };
};

void say(char *msg)
{
  if(!send_com_say(&(client.sock),msg)) logmsg("error sending say command.!!");
  else {
    if (LOGIT && LOG_SEND) {
      fprintf(client.log.file, "\nCommand send in time %d :", world.time.clock);
      fprintf(client.log.file, "\n  Saying : %s\n", msg);
    };
    world.body.plcount.say++;
  };
};

void eye(bool state)
{
  if(!send_com_eye(&(client.sock),state)) logmsg("error sending eye command.!!");
  else {
    if (LOGIT && LOG_SEND) {
      fprintf(client.log.file, "\nCommand send in time %d :", world.time.clock);
      fprintf(client.log.file, "\n  Eye State : %s\n", state?"ON":"OFF");
    };
  };
}

bool perform_command(Command &com)
{
  bool result = true;

  switch (com.type) {
    case NO_COMMAND:
    break;
    case KICK:
      kick(com.par1,com.par2);
    break;
    case DASH:
      dash(com.par1);
    break;
    case TURN:
      turn(com.par1);
    break;
  };

  return result;
}

void say_world_model(void)
{
  int i;
  char s[MAX_SAY_SIZE], st[MAX_SAY_SIZE];

  sprintf(s,"%s%s%d %d %d ",TEAM_SIGNATURE, client.teamname, player.unum, world.time.clock, (int)world.body.stamina);
  sprintf(st,"*%d %.2f %.2f %.2f %.2f ", (int)world.ball.correctness, world.ball.x, world.ball.y, world.ball.vx, world.ball.vy);
  strcat(s,st);
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) 
    if (world.tmate[i].correctness>MIN_USEFUL_CRR) {
      sprintf(st,"*%d %d %.2f %.2f ", i, (int)world.tmate[i].correctness, world.tmate[i].x, world.tmate[i].y);
      strcat(s,st);
    };
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) 
    if (world.opp[i].correctness>MIN_USEFUL_CRR) {
      sprintf(st,"*-%d %d %.2f %.2f ", i, (int)world.opp[i].correctness, world.opp[i].x, world.opp[i].y);
      strcat(s,st);
    };
//  if (sensor.comm.sth2say) {
    if (sensor.comm.passdest!=0)
      sprintf(st,"*12 %d 0 0", (int)sensor.comm.passdest);
      strcat(s,st);
//  }
  
  say(s);
  
  sensor.comm.sth2say = false;
  sensor.comm.passdest = 0;
}

void coach_say_world_model(void)
{
  int i;
  char s[MAX_SAY_SIZE], st[MAX_SAY_SIZE];

  sprintf(s,"%.1f %.1f ", world.ball.x, world.ball.y);

  for(i=1;i<=N_PLAYER_PER_TEAM;i++) {
    sprintf(st,"%d %d ", (int)world.tmate[i].x, (int)world.tmate[i].y);
    strcat(s,st);
  };
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) {
    sprintf(st,"%d %d ", (int)world.opp[i].x, (int)world.opp[i].y);
    strcat(s,st);
  };
  strcpy(st,s);
  sprintf(s, "(freeform \"%s\")", st);
  
  say(s);
  
  sensor.comm.sth2say = false;
  sensor.comm.passdest = 0;
}

void neck2angle(double dir)
{
  double neck_ang = normalize_angle(dir);

  if (sensor.just_sent_command.type == TURN)
    neck_ang = normalize_angle(neck_ang - sensor.just_sent_command.par1);

  if (neck_ang>90) neck_ang=90;
  if (neck_ang<-90) neck_ang=-90;
  if (fabs(neck_ang-world.body.head_angle)>10)
    turn_neck(normalize_angle(neck_ang-world.body.head_angle));
}

void neck2ball(void)
{
  double neck_ang;
  neck_ang=reldir_2me(estimate_ball_x(1),estimate_ball_y(1));
  neck2angle(neck_ang);
}

void neck2zero(void)
{
  if (fabs(world.body.head_angle)>10)
    turn_neck(-world.body.head_angle);
}
